/*******************************************************************************
 * Reference:
 * https://github.com/quarkslab/llvm-dev-meeting-tutorial-2015/tree/master/MBA *
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/Transforms/IPO/PassManagerBuilder.h"
#include "llvm/IR/Instruction.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"

using namespace llvm;

namespace {
  class Obfuscate : public BasicBlockPass {
    public:
      static char ID;
      Obfuscate() : BasicBlockPass(ID) {}

      bool runOnBasicBlock(BasicBlock &BB) override {
        /* Will need to return whether or not the function was modified. */
        bool modified = false;

        /* Iterate through all the instructions in this basic block. */
        for (auto it = BB.begin(); it != BB.end(); ++it) {
          Instruction &inst = *it;

          /* Check if the instruction represents and addition. */
          if (inst.getOpcode() == Instruction::Add) {
            auto *binaryOp = dyn_cast<BinaryOperator>(&inst);
            if (!binaryOp) {
              continue;
            }

            /* Check if the operands are integers. */
            if (binaryOp->getType()->isIntegerTy()) {
              /* Create the new instruction. */
              IRBuilder<> Builder(binaryOp);

              /* a + b = (a ^ b) + ((a & b) << 1) */
              Value *obfuscated = Builder.CreateAdd(
                  Builder.CreateXor(
                    binaryOp->getOperand(0),
                    binaryOp->getOperand(1)),
                  Builder.CreateShl(
                    Builder.CreateAnd(
                      binaryOp->getOperand(0),
                      binaryOp->getOperand(1)),
                    ConstantInt::get(binaryOp->getType(), 1)));

              /* Replace the current instruction with the newly created one. */
              ReplaceInstWithValue(BB.getInstList(), it, obfuscated);
              modified = true;
            }
          }
        }

        return modified;
      }
  };
}

char Obfuscate::ID = 0;

static void registerObfuscatePass(const PassManagerBuilder &,
    legacy::PassManagerBase &PM) {
  PM.add(new Obfuscate());
}

/* Enable pass on Level 0 optimization. */
static RegisterStandardPasses
    RegisterMyPass0(PassManagerBuilder::EP_EnabledOnOptLevel0, registerObfuscatePass);
